package uk.ac.soton.grophysics;

import javax.vecmath.Matrix3d;
import javax.vecmath.Matrix4d;
import javax.vecmath.Matrix4f;
import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;

import com.bulletphysics.dynamics.constraintsolver.Generic6DofConstraint;
import com.bulletphysics.linearmath.Transform;

import de.grogra.graph.impl.Node;
import de.grogra.imp3d.objects.Cylinder;
import de.grogra.rgg.Library;

public class FlexNode extends Cylinder {
	/**
	 * random id generated by eclipse
	 */
	private static final long serialVersionUID = -7688781891937691880L;

	static float DEFAULT_MASS = 0.001f; 	// 1g 
	static float DEFAULT_STIFFNESS = 0.5f;  // stiffness range 0 - 1 

	public float mass; // mass in kg (located at the centre of gravity)
    //enh:field getter setter

	public float stiffnessT; // Translational stiffness of the joint
    //enh:field getter setter

	public float stiffnessX; // Rotational stiffness of the joint about X
    //enh:field getter setter

	public float stiffnessY; // Rotational stiffness of the joint about Y
    //enh:field getter setter

	public float stiffnessZ; // Rotational stiffness of the joint about Z
    //enh:field getter setter

	public float rotDamping; // Rotational Joint damping
    //enh:field getter setter

	public float linDamping; // Linear Joint damping
    //enh:field getter setter

	public float linBounce; // Linear bounciness (restitution)
    //enh:field getter setter

	public float rotBounce; // Rotational bounciness (restitution)
    //enh:field getter setter

	
	// the joint is at the bottom of the stem segment
	Generic6DofConstraint joint;
	// transforms the bottom joint frame to change the idle position of the joint
	Matrix4d jointFrameOffset;

		public FlexNode() {
		this(1f, 0.05f, DEFAULT_MASS);
	}

	public FlexNode(float length, float radius) {
		this(length, radius, DEFAULT_MASS);
	}

	public FlexNode(float length, float radius, float mass) {
		super(length, radius);
		this.mass = mass;
		this.stiffnessT = DEFAULT_STIFFNESS;
		this.stiffnessX = DEFAULT_STIFFNESS;
		this.stiffnessY = DEFAULT_STIFFNESS;
		this.stiffnessZ = DEFAULT_STIFFNESS;
		this.linDamping = 1.0f;
		this.rotDamping = 1.0f;
		this.linBounce = 0.5f;
		this.rotBounce = 0f;
		jointFrameOffset = new Matrix4d();
		jointFrameOffset.setIdentity();
//		Library.println("New FlexNode mass = "+mass);
	}

	public float getStiffness(int i) {
		float s = 0;
		switch (i) {
			case 0: {
				s = stiffnessX; 
				break;
			}
			case 1: {
				s = stiffnessY; 
				break;
			}
			case 2: {
				s = stiffnessZ; 
				break;
			}
			case 3: {
				s = stiffnessT; 
				break;
			}
		}
		return s;
	}
	
	public Generic6DofConstraint getJoint() {
		return joint;
	}

	public void setJoint(Generic6DofConstraint joint) {
		this.joint = joint;
	}

	/**
	 * adds a proportion of the joint displacement back into the joint frame to fix the 
	 * angle as the stem ages
	 * @param rate
	 */
	public void ageJoint(double rate) {
		// print the matrix
//		Library.println("AgeJoint: jointFrameOffset before "+jointFrameOffset);
		if (joint == null) {
			Library.println("WARNING: FlexNode.ageJoint(): joint is NULL");
		} else {
			Matrix3d mat = new Matrix3d();
			Matrix3d relative_frame = new Matrix3d();
			relative_frame.setIdentity();
			// 
			double angleX = joint.getAngle(0);
			double angleY = joint.getAngle(1);
			double angleZ = joint.getAngle(2);
//			Library.println("AgeJoint: angles XYZ "+angleX+", "+angleY+", "+angleZ);
			
			// reduce the angles according to the rate
			angleX *= rate;
			angleY *= rate;
			angleZ *= rate;
//			Library.println("AgeJoint: scaled angles XYZ "+angleX+", "+angleY+", "+angleZ);
			// create a matrix from the rotation
			mat.rotX(angleX);
//			Library.println("AgeJoint: X rotation "+mat);
			relative_frame.mul(mat);
			mat.rotY(angleY);
//			Library.println("AgeJoint: Y rotation "+mat);
			relative_frame.mul(mat);
			mat.rotZ(angleZ);
//			Library.println("AgeJoint: Z rotation "+mat);
			relative_frame.mul(mat);
			
//			Library.println("AgeJoint: scaled matrix "+relative_frame);
			// get the rotation as a transform
			Matrix4d rotate = new Matrix4d();
			rotate.setIdentity();
			rotate.set(relative_frame);

			// update the joint frame
			jointFrameOffset.mul(rotate);
		}		
//		Library.println("AgeJoint: jointFrameOffset after "+jointFrameOffset);
	}
	
	/**
	 * returns a transform corresponding to the initial
	 * bottom joint frame relative to the centre of mass
	 * NB this is the frame when the joint is NOT under tension
	 * @return Transform
	 */
	public Transform initBottomJointFrame() {
		// get a new transform
		Transform jointFrame = new Transform();
		jointFrame.setIdentity();
		// translate the origin to the joint frame
		Vector3f anchor = new Vector3f(0, 0, -length/2);
		jointFrame.origin.set(anchor);	
		return jointFrame;
	}

	/**
	 * returns a transform corresponding to the 
	 * bottom joint frame relative to the centre of mass
	 * NB this is the frame when the joint is NOT under tension
	 * @return Transform
	 */
	public Transform getBottomJointFrame() {
		// update the frame because the length may have changed
		Transform jointFrame = initBottomJointFrame();
		// multiply by the rotational offset matrix

		Matrix4f jf= new Matrix4f();
		jointFrame.getMatrix(jf);
    	Matrix4d jfd = new Matrix4d(jf);
		
		jfd.mul(jointFrameOffset);
		jf.set(jfd);
		jointFrame.set(jf);
//		Library.println("FlexNode: getBottomJointFrame = "+jf);
		return jointFrame;
	}

	Vector3d getLengthVector() {
		return new Vector3d( 0, 0, length);
	}
	
	/**
	 * returns a transform corresponding to the 
	 * top joint position relative to the centre of mass
	 * @return Transfrom
	 */
	public Transform getTopJointFrame() {
		// get a new transform
		Transform jointFrame = new Transform();
		jointFrame.setIdentity();
		// translate the origin to the joint frame
		Vector3f anchor = new Vector3f(0, 0, length/2);
		jointFrame.origin.set(anchor);	
		return jointFrame;
	}
	
	/**
	 * Find the previous FlexNode in the current branch
	 * returns null if nothing is found
	 * @return FlexNode
	 */
	public FlexNode getPreviousFlexNode() {
		FlexNode previous = null;
		Node pred = this.getPredecessor();
		while (pred != null) {
			if (pred instanceof FlexNode) {
				previous = (FlexNode) pred;
				break;
			}
			pred = pred.getPredecessor();
		}
		return previous;
	}

	/**
	 * Find the previous FlexNode in the tree by searching the branch first, 
	 * then hierarchically through it's parents
	 * returns null if nothing is found
	 * @return FlexNode
	 */
	public FlexNode getParentFlexNode() {
		FlexNode parentFlexNode = getPreviousFlexNode();
		if  (parentFlexNode == null) {
			// no more FlexNodes in this branch so look for a parent node
			Node parent = this.getAxisParent();
find:		while (parent != null) {
				if (parent instanceof FlexNode) {
					parentFlexNode = (FlexNode) parent;
					break find;
				} else {
					// try the predecessors of the parent
					Node pred = parent.getPredecessor();
					while (pred != null) {
						if (pred instanceof FlexNode) {
							parentFlexNode = (FlexNode) pred;
							break find;
						}
						pred = pred.getPredecessor();
					}
				}
				// try the previous parent
				parent = parent.getAxisParent();
			}
		}
		return parentFlexNode;
	}
	

//enh:insert
//enh:begin
// NOTE: The following lines up to enh:end were generated automatically

	public static final NType $TYPE;

	public static final NType.Field mass$FIELD;
	public static final NType.Field stiffnessT$FIELD;
	public static final NType.Field stiffnessX$FIELD;
	public static final NType.Field stiffnessY$FIELD;
	public static final NType.Field stiffnessZ$FIELD;
	public static final NType.Field rotDamping$FIELD;
	public static final NType.Field linDamping$FIELD;
	public static final NType.Field linBounce$FIELD;
	public static final NType.Field rotBounce$FIELD;

	private static final class _Field extends NType.Field
	{
		private final int id;

		_Field (String name, int modifiers, de.grogra.reflect.Type type, de.grogra.reflect.Type componentType, int id)
		{
			super (FlexNode.$TYPE, name, modifiers, type, componentType);
			this.id = id;
		}

		@Override
		public void setFloat (Object o, float value)
		{
			switch (id)
			{
				case 0:
					((FlexNode) o).mass = (float) value;
					return;
				case 1:
					((FlexNode) o).stiffnessT = (float) value;
					return;
				case 2:
					((FlexNode) o).stiffnessX = (float) value;
					return;
				case 3:
					((FlexNode) o).stiffnessY = (float) value;
					return;
				case 4:
					((FlexNode) o).stiffnessZ = (float) value;
					return;
				case 5:
					((FlexNode) o).rotDamping = (float) value;
					return;
				case 6:
					((FlexNode) o).linDamping = (float) value;
					return;
				case 7:
					((FlexNode) o).linBounce = (float) value;
					return;
				case 8:
					((FlexNode) o).rotBounce = (float) value;
					return;
			}
			super.setFloat (o, value);
		}

		@Override
		public float getFloat (Object o)
		{
			switch (id)
			{
				case 0:
					return ((FlexNode) o).getMass ();
				case 1:
					return ((FlexNode) o).getStiffnessT ();
				case 2:
					return ((FlexNode) o).getStiffnessX ();
				case 3:
					return ((FlexNode) o).getStiffnessY ();
				case 4:
					return ((FlexNode) o).getStiffnessZ ();
				case 5:
					return ((FlexNode) o).getRotDamping ();
				case 6:
					return ((FlexNode) o).getLinDamping ();
				case 7:
					return ((FlexNode) o).getLinBounce ();
				case 8:
					return ((FlexNode) o).getRotBounce ();
			}
			return super.getFloat (o);
		}
	}

	static
	{
		$TYPE = new NType (new FlexNode ());
		$TYPE.addManagedField (mass$FIELD = new _Field ("mass", _Field.PUBLIC  | _Field.SCO, de.grogra.reflect.Type.FLOAT, null, 0));
		$TYPE.addManagedField (stiffnessT$FIELD = new _Field ("stiffnessT", _Field.PUBLIC  | _Field.SCO, de.grogra.reflect.Type.FLOAT, null, 1));
		$TYPE.addManagedField (stiffnessX$FIELD = new _Field ("stiffnessX", _Field.PUBLIC  | _Field.SCO, de.grogra.reflect.Type.FLOAT, null, 2));
		$TYPE.addManagedField (stiffnessY$FIELD = new _Field ("stiffnessY", _Field.PUBLIC  | _Field.SCO, de.grogra.reflect.Type.FLOAT, null, 3));
		$TYPE.addManagedField (stiffnessZ$FIELD = new _Field ("stiffnessZ", _Field.PUBLIC  | _Field.SCO, de.grogra.reflect.Type.FLOAT, null, 4));
		$TYPE.addManagedField (rotDamping$FIELD = new _Field ("rotDamping", _Field.PUBLIC  | _Field.SCO, de.grogra.reflect.Type.FLOAT, null, 5));
		$TYPE.addManagedField (linDamping$FIELD = new _Field ("linDamping", _Field.PUBLIC  | _Field.SCO, de.grogra.reflect.Type.FLOAT, null, 6));
		$TYPE.addManagedField (linBounce$FIELD = new _Field ("linBounce", _Field.PUBLIC  | _Field.SCO, de.grogra.reflect.Type.FLOAT, null, 7));
		$TYPE.addManagedField (rotBounce$FIELD = new _Field ("rotBounce", _Field.PUBLIC  | _Field.SCO, de.grogra.reflect.Type.FLOAT, null, 8));
		$TYPE.validate ();
	}

	@Override
	protected NType getNTypeImpl ()
	{
		return $TYPE;
	}

	@Override
	protected de.grogra.graph.impl.Node newInstance ()
	{
		return new FlexNode ();
	}

	public float getMass ()
	{
		return mass;
	}

	public void setMass (float value)
	{
		this.mass = (float) value;
	}

	public float getStiffnessT ()
	{
		return stiffnessT;
	}

	public void setStiffnessT (float value)
	{
		this.stiffnessT = (float) value;
	}

	public float getStiffnessX ()
	{
		return stiffnessX;
	}

	public void setStiffnessX (float value)
	{
		this.stiffnessX = (float) value;
	}

	public float getStiffnessY ()
	{
		return stiffnessY;
	}

	public void setStiffnessY (float value)
	{
		this.stiffnessY = (float) value;
	}

	public float getStiffnessZ ()
	{
		return stiffnessZ;
	}

	public void setStiffnessZ (float value)
	{
		this.stiffnessZ = (float) value;
	}

	public float getRotDamping ()
	{
		return rotDamping;
	}

	public void setRotDamping (float value)
	{
		this.rotDamping = (float) value;
	}

	public float getLinDamping ()
	{
		return linDamping;
	}

	public void setLinDamping (float value)
	{
		this.linDamping = (float) value;
	}

	public float getLinBounce ()
	{
		return linBounce;
	}

	public void setLinBounce (float value)
	{
		this.linBounce = (float) value;
	}

	public float getRotBounce ()
	{
		return rotBounce;
	}

	public void setRotBounce (float value)
	{
		this.rotBounce = (float) value;
	}

//enh:end

}
